import concurrent.futures
import logging
import os
import socket
import warnings
from collections import OrderedDict, defaultdict
from configparser import NoOptionError
from itertools import islice
from multiprocessing import Pool
from numbers import Number
from operator import getitem
from typing import TYPE_CHECKING, Optional, Tuple
from urllib.parse import urljoin, urlparse

if TYPE_CHECKING:
    from urllib.parse import ParseResult

import requests
from elasticsearch.exceptions import ElasticsearchDeprecationWarning
from flask import jsonify
from jinja2 import Template
from requests import ConnectionError, Timeout

from squirro.common.config import get_config
from squirro.common.dependency import get_injected
from squirro.common.es_utils import connect_elastic_from_config
from squirro.common.tracing.elasticsearch_tracing import TracingTransport
from squirro.sdk.studio import StudioPlugin
from squirro_client import SquirroClient

json = get_injected("json")

plugin = StudioPlugin(__name__)
log = logging.getLogger(__name__)
config = get_injected("config")

# We setup the config for Topic API to get also ES related settings.
config = get_config("squirro.api.topic")

# Tracking the Squirro Release number
VERSION = ""

logging.basicConfig(level=logging.INFO, format="%(message)s", datefmt="%m-%d %H:%M")

# set elasticsearch log level to warning
es_logger = logging.getLogger("elasticsearch")
es_logger.setLevel(logging.WARNING)

# ignore warnings about deprecations
warnings.filterwarnings("ignore", category=ElasticsearchDeprecationWarning)


KNOWN_SERVICES = [
    "cluster",
    "configuration",
    "content",
    "datasource",
    "digestmailer",
    "emailsender",
    "filtering",
    "fingerprint",
    "ingester",
    "machinelearning",
    "notes",
    "pdfconversion",
    "plumber",
    "provider",
    "relatedstory",
    "scheduler",
    "search",
    "thumbler",
    "topic",
    "topicproxy",
    "trends",
    "user",
    "userproxy",
    "webshot",
]

AVAILABLE_SERVICES = []
CONFIG_MAP = {}

for service in KNOWN_SERVICES:
    config_file = os.path.join("/etc/squirro", f"{service}.ini")

    if os.path.exists(config_file):
        AVAILABLE_SERVICES.append(service)
    else:
        continue

    try:
        CONFIG_MAP[service] = config.get("services", service)
    except NoOptionError:
        pass


def check_elasticsearch():
    try:
        es = connect_elastic_from_config(config, transport_class=TracingTransport)
        return es.cluster.health()["status"] in ("green", "yellow")
    except Exception:
        return False


def check_service(service):
    status_url = f"{CONFIG_MAP[service]}/_internal/status"
    log.info(f"Checking {service} at {status_url}...")

    try:
        response = requests.get(status_url, timeout=20)
    except requests.exceptions.Timeout:
        return service

    if response.status_code != 200:
        return service

    return None


def request_es(path):
    host = config.get("index", "es_index_servers").split(",")[0]

    # make sure urljoin will not strip part of the path
    if not host.endswith("/"):
        host = f"{host}/"

    try:
        user = config.get("index", "es_auth_user")
        password = config.get("index", "es_auth_password")
        auth = (user, password)
    except NoOptionError:
        auth = None

    try:
        verify = config.getboolean("index", "es_verify_certs")
    except NoOptionError:
        verify = False

    r = requests.get(
        urljoin(host, path),
        auth=auth,
        verify=verify,
        headers={"Content-Type": "application/json"},
    )

    return r.json()


@plugin.route("/healthcheck", methods=["get"], skip_authentication=True)
def healthcheck():
    """
    Checks the status of all available services.

    Returns:
        A Flask response with status 200 OK if all services are fine, or status 500 with a list of all failed services.
    """
    failed_services = []
    with concurrent.futures.ThreadPoolExecutor() as executor:
        results = executor.map(check_service, AVAILABLE_SERVICES)
        for service in results:
            if service:
                failed_services.append(service)

    if not check_elasticsearch():
        failed_services.append("elasticsearch")

    if failed_services:
        return (
            f"Services not fine: {', '.join(failed_services)}",
            200,
        )
    else:
        return "OK", 200


@plugin.route("/status", methods=["get"])
def status():
    client = get_injected("squirro_client")
    redis_client = get_injected("redis_studio")
    prefix = redis_prefix(client)

    template_file = os.path.join(os.path.dirname(__file__), "template.html")
    with open(template_file) as f:
        template = Template(f.read())

        cluster_info = get_cluster_info()
        # which host is this?
        hostname = socket.getfqdn().lower()

        # figure out what services to call in paralell
        p = Pool(4)
        responses = p.map(get_service_url, KNOWN_SERVICES)
        p.close()
        p.join()

        # get status calls in paralell
        status_params = []
        for service, parsed_url in responses:
            # service installed?
            if not parsed_url:
                continue

            for member in cluster_info["members"]:
                member = member.lower()

                if hostname != member:
                    parsed_url = parsed_url._replace(netloc=f"{member}:81")

                # construct the URL for the status endpoint of the service
                status_url = parsed_url._replace(
                    path=os.path.join(parsed_url.path, "_internal/status")
                ).geturl()

                params = {
                    "node": member,
                    "url": status_url,
                    "service": service,
                }

                status_params.append(params)

        pool_size = len(cluster_info["members"]) * 3
        # cap paralell calls at 10
        if pool_size > 10:
            pool_size = 10

        p = Pool(pool_size)
        responses = p.map(get_service_status, status_params)
        p.close()
        p.join()

        # building status_dict from responses
        participants = defaultdict(dict)

        for node, service, state, message in responses:
            if "services" not in participants[node]:
                participants[node]["services"] = {}

            if "errors" not in participants[node]:
                participants[node]["errors"] = 0

            if "warnings" not in participants[node]:
                participants[node]["warnings"] = 0

            if "warnings" not in participants[node]:
                participants[node]["message"] = ""

            if state == "error":
                icon = "error"
                participants[node]["errors"] += 1
            elif state == "warn":
                icon = "info"
                participants[node]["warnings"] += 1
            else:
                icon = "check_circle"

            participants[node]["services"][service] = {
                "state": state,
                "message": message,
                "icon": icon,
            }

        # os stats currently only for the local machine
        participants[hostname]["stats"] = os_stats()

        # Elasticsearch Health Status
        cluster_health = request_es("_cluster/health")

        # Elasticsearch settings
        nodes = request_es("_nodes/jvm,settings")

        # Elasticsearch stats
        stats = request_es("_nodes/stats/jvm,os,fs,process")

        # merge settings standby stats
        for node, node_stats in stats["nodes"].items():
            for key, value in node_stats.items():
                if key in nodes["nodes"][node]:
                    stats_key = f"{key}_stats"
                    nodes["nodes"][node][stats_key] = value
                else:
                    nodes["nodes"][node][key] = value

        # Elasticsearch indexes
        indexes = request_es("_cat/indices?format=json")

        # translatable indexes are on top
        known_indexes = []
        unknown_indexes = []

        # translate index names
        for index in indexes:
            if index["pri"] and index["rep"]:
                index["shards"] = f"{index['pri']} x {int(index['rep']) + 1}"
            else:
                index["shards"] = "-"

            key = "index_{}".format(index["index"])
            project = redis_get(redis_client, prefix, key)

            if project:
                index["project_name"] = project["name"]
                index["owner_name"] = project["owner_name"]
                index["owner_email"] = project["owner_email"]

                known_indexes.append(index)
            else:
                index["project_name"] = index["index"]
                index["owner_name"] = "unknown"
                index["owner_email"] = "unknown"
                unknown_indexes.append(index)

        indexes = known_indexes + unknown_indexes

        out = template.render(
            cluster_mode=cluster_info["mode"],
            cluster_leader=cluster_info["leader"],
            participants=participants,
            cluster_health=cluster_health,
            nodes=nodes,
            indexes=indexes,
        )
    return out


@plugin.route("/ingester-status-html", methods=["get"])
def ingester_status_html():
    client = get_injected("squirro_client")
    cluster_info = get_cluster_info()
    hostname = socket.getfqdn().lower()

    totals = defaultdict(int)
    nodes = OrderedDict()
    projects = {}
    sources = {}

    for member in cluster_info["members"]:
        member = member.lower()
        if hostname == member:
            # Cluster needs to be `http://nginx` for local Docker development environment
            cluster = "http://nginx" if in_docker_container() else "http://localhost:81"
        else:
            cluster = f"http://{member}:81"

        log.debug(f"Using `{cluster}` as a cluster")

        ingester_client = SquirroClient(
            client.client_id, client.client_secret, cluster=cluster
        )

        try:
            ingester_client.authenticate(refresh_token=client._refresh_token)
            pipeline_data = ingester_client.get_pipeline_status(
                sections=["backlog_total", "errors_total"]
            )
            projects_data = ingester_client.get_projects_status()
            sources_data = ingester_client.get_sources_status()
        except Exception:
            node = {"items": "error", "batches": "error", "failed_batches": "error"}

            nodes[member] = node
            continue

        items = pipeline_data["backlog_total"]["items"]
        batches = pipeline_data["backlog_total"]["batches"]

        failed_batches = 0

        for sub, count in pipeline_data["errors_total"].items():
            failed_batches += count["batches"]

        node = {"items": items, "batches": batches, "failed_batches": failed_batches}

        nodes[member] = node
        totals["items"] += items
        totals["batches"] += batches
        totals["failed_batches"] += failed_batches

        projects = sum_up_nested_dictionary_values(
            from_dictionary=projects_data, to_dictionary=projects
        )
        sources = sum_up_nested_dictionary_values(
            from_dictionary=sources_data, to_dictionary=sources
        )

    sorted_projects = sort_entity_data_per_key(entity_data=projects, key="batches")
    sorted_sources = sort_entity_data_per_key(entity_data=sources, key="batches")

    template_file = os.path.join(os.path.dirname(__file__), "template_ingester.html")
    with open(template_file) as f:
        template = Template(f.read())

    ret = template.render(
        totals=totals,
        nodes=nodes,
        projects=slice_ordered_dictionary(
            ordered_dict=sorted_projects, start=0, stop=10
        ),
        sources=sorted_sources,
    )

    return jsonify(html=ret)


@plugin.route("/ingester-status", methods=["get"])
def ingester_status():
    # ingester
    client = get_injected("squirro_client")
    redis_client = get_injected("redis_studio")
    prefix = redis_prefix(client)
    pipeline_data = client.get_pipeline_status(
        sections=["backlog_total", "errors_total", "backlog_detailed"]
    )
    backlog_detailed = []

    # expand detailed backlog with metadata
    for source_id, stats in pipeline_data["backlog_detailed"].items():
        key = f"source_{source_id}"
        source_metadata = redis_get(redis_client, prefix, key)

        if not source_metadata:
            source_metadata = {}

        detail = {
            "source_id": source_id,
            "source_name": source_metadata.get("title", source_id),
            "project_name": source_metadata.get("project_name", "unknown"),
            "workflow_name": source_metadata.get("workflow_name", "unknown"),
            "owner_name": source_metadata.get("owner_name", "unknown"),
            "owner_email": source_metadata.get("owner_email", "unknown"),
            "batches": stats["batches"],
            "items": stats["items"],
        }

        backlog_detailed.append(detail)

    pipeline_data["backlog_detailed"] = backlog_detailed

    return jsonify(pipeline_data=pipeline_data)


@plugin.route("/metadata", methods=["get"])
def metadata():
    client = get_injected("squirro_client")
    redis_client = get_injected("redis_studio")

    index_key_count = 0
    project_key_count = 0
    source_key_count = 0
    failed_projects = []
    transactions = 0
    prefix = redis_prefix(client)

    # clear all keys
    keys = redis_client.keys(f"{prefix}*")

    # get all squirro projects
    projects = client.get_projects()
    transactions += 1

    for project in projects:
        project_data = client.get_project(project_id=project["id"])
        transactions += 1

        # get project owner
        members = client.get_project_members(project["id"])
        transactions += 1

        owner_email = "unknown"
        owner_name = "unknown"

        for member in members["members"]:
            if member["project_role"] == "owner":
                owner_email = member.get("email")
                owner_name = member.get("full_name")

                if not owner_name:
                    owner_name = owner_email

                break

        # get all sources and store lookup data
        try:
            sources = client.get_sources(project["id"])
        except Exception:
            failed_projects.append(project["id"])
            continue

        transactions += 1

        for source in sources:
            key = f"source_{source['id']}"

            try:
                workflow_name = client.get_pipeline_workflow(
                    project["id"], source["pipeline_workflow_id"]
                )
            except Exception:
                log.exception(
                    "Can not get workflow %r in project %r",
                    source["pipeline_worklow_id"],
                    project["id"],
                )
                workflow_name = "unknown"

            data = {
                "title": source["name"],
                "workflow_id": source["pipeline_workflow_id"],
                "workflow_name": workflow_name,
                "project_id": project["id"],
                "project_name": project_data["title"],
                "owner_name": owner_name,
                "owner_email": owner_email,
            }

            redis_set(redis_client, prefix, key, data)
            source_key_count += 1

        # store project data
        project_meta = {
            "id": project["id"],
            "name": project_data["title"],
            "owner_name": owner_name,
            "owner_email": owner_email,
        }

        key = f"project_{project['id']}"
        redis_set(redis_client, prefix, key, project_meta)
        project_key_count += 1

        # store elastic index lookup
        locator = project_data.get("locator")
        index_names = []

        if locator:
            index_names = locator.get("index_names")

        if not index_names:
            index_names = []

        for index_name in index_names:
            key = f"index_{index_name}"
            redis_set(redis_client, prefix, key, project_meta)
            index_key_count += 1

    metadata = {}
    metadata["tenant"] = client.tenant
    metadata["prefix"] = prefix
    metadata["deleted_keys"] = len(keys)
    metadata["project_key_count"] = project_key_count
    metadata["source_key_count"] = source_key_count
    metadata["index_key_count"] = index_key_count
    metadata["transactions"] = transactions
    metadata["failed_projects"] = failed_projects

    return jsonify(metadata=metadata)


def redis_set(client, prefix, key, value):
    client.set(redis_key(prefix, key), json.dumps(value))


def redis_get(client, prefix, key):
    data = client.get(redis_key(prefix, key))
    if data:
        ret = json.loads(data)
    else:
        ret = None

    return ret


def redis_prefix(client):
    return f"sq_monitoring_{client.tenant}"


def redis_key(prefix, key):
    return f"{prefix}_{key}"


def os_stats():
    ret = {}
    ret["uptime"] = os.popen(
        'awk \'{print int($1/86400)" days "int($1%86400/3600)":"int(($1%3600)/60)":"int($1%60)}\' /proc/uptime'
    ).read()  # noqa
    ret["ram_usage"] = "{:.2f}".format(
        float(os.popen("free | grep Mem | awk '{print $3/$2 * 100.0}'").read())
    )
    ret["disk_usage_var"] = os.popen("df /var | awk 'END{print $4}'").read()
    ret["disk_usage_opt"] = os.popen("df /opt | awk 'END{print $4}'").read()
    ret["disk_usage_varlog"] = os.popen("df /var/log | awk 'END{print $4}'").read()
    ret["server_load"] = (
        os.popen("uptime | awk '/load average/ {printf(\"%s\",($10 $11 $12))}'")
        .read()
        .replace(",", " | ")
    )
    ret["server_date"] = os.popen("date").read()

    return ret


def get_service_url(service: str) -> Tuple[str, Optional["ParseResult"]]:
    """
    Detects if a service is installed and what its URL is.
    :param service: the name of the service (e.g. datasource)
    """

    config = get_injected("config")
    config_file = f"/etc/squirro/{service}.ini"

    if not os.path.exists(config_file):
        return (service, None)

    try:
        url = config.get("services", service)
    except NoOptionError:
        log.warning("Failed to find service url for %r in config", service)
        return (service, None)

    if not url:
        log.warning("Failed to find service url for %r", service)
        return (service, None)

    try:
        parsed_url = urlparse(url)
    except ValueError:
        log.warning("Failed to parse the url for %r", service)
        return service, None

    return service, parsed_url


def get_service_status(params: dict) -> Tuple[str, str, str, str]:
    node = params["node"]
    service = params["service"]
    url = params["url"]

    try:
        r = requests.get(url)
    except ConnectionError:
        return (
            node,
            service,
            "warn",
            "Failed to connect. Nginx down? Firewall blocking?",
        )
    except Timeout:
        return (node, service, "warn", "Connection timed out. High server load?")
    except Exception:
        log.error(f"Unknown error happened when calling {url}")
        return (
            node,
            service,
            "error",
            f"Unknown error happened when calling {url}",
        )

    if r.status_code == 200:
        if "OK" in r.text:
            return (node, service, "ok", r.text)
        else:
            return (node, service, "warn", r.text)

    if r.status_code == 403:
        return (node, service, "warn", "Not allowed to connect to service status")

    if r.status_code == 404:
        return (node, service, "warn", "404 - Not found")

    if r.status_code == 500:
        return (
            node,
            service,
            "error",
            "500 - Unknown service error, check /var/log/squirro/{0}/{0}.log and /var/log/squirro/{0}/stderr.log".format(
                service
            ),
        )  # noqa

    if r.status_code == 502:
        return (
            node,
            service,
            "error",
            f"502 - Bad gateway, run service sq{service}d status to check",
        )

    return (node, service, "warn", f"{r.status_code} - {r.text}")


def get_cluster_info() -> dict:
    config = get_injected("config")

    ret = {"members": []}

    discovery = config.get("redis", "endpoint_discovery")
    if not discovery or discovery.lower() == "false":
        ret["mode"] = "standalone"
        ret["leader"] = socket.getfqdn()
        participants = [socket.getfqdn()]
    else:
        r = requests.get("http://localhost:81/service/cluster/v0/leader/cluster.json")
        if r.status_code == 200:
            resp = r.json()
            ret["mode"] = resp["data"]["myState"]
            ret["leader"] = resp["data"]["electionLeader"]
            participants = resp["data"]["electionParticipants"]

    participants.sort()

    for participant in participants:
        ret["members"].append(urlparse(f"http://{participant}").hostname)

    return ret


def in_docker_container() -> bool:
    """Return true if we run in a docker container, otherwise false."""
    if os.path.isfile("/.dockerenv"):
        return True
    return False


# TODO: [SQ-16527] Move to squirro.common
def sum_up_nested_dictionary_values(from_dictionary: dict, to_dictionary: dict) -> dict:
    """Sum up nested integer dictionary values.

    :param from_dictionary: dict, data to add from
    :param to_dictionary: dict, data to add to (will be mutated and returned)
    :return: dict, data with summed up nested dictionary values
    """
    for key, values in from_dictionary.items():
        to_values = to_dictionary.setdefault(key, {})

        for value_key, value_value in values.items():
            if isinstance(value_value, Number):
                to_values[value_key] = to_values.get(value_key, 0) + value_value
            else:
                to_values[value_key] = value_value

    return to_dictionary


def sort_entity_data_per_key(
    entity_data: dict, key: str, reverse: Optional[bool] = True
) -> OrderedDict:
    """Based on the provided entity data dictionary returns OrderedDict sorted per key.

    :param entity_data: dict, data to be sorted
    :param key: str, key to be sorted for
    :param reverse: Optional[bool], flag to sort in descending order, default: True
    :return: OrderedDict, data sorted per key
    """
    return OrderedDict(
        sorted(
            entity_data.items(), key=lambda item: getitem(item[1], key), reverse=reverse
        )
    )


def slice_ordered_dictionary(
    ordered_dict: OrderedDict, start: int, stop: int
) -> OrderedDict:
    """Return slice from the provided OrderedDict.

    :param ordered_dict: OrderedDict, dictionary to slice
    :param start: int, index where the slicing starts
    :param stop: int, index where the slicing stops
    :return: OrderedDict, sliced dictionary
    """
    return OrderedDict(islice(ordered_dict.items(), start, stop))
